--*******************************************************************************
-- Actel Corportation
-- Fusion's Voltage Comparator Application Note's Design Example
--
-- File Name:    TOP_CM.vhd
-- Date:         December 2006
-- Language:     VHDL
--
--*******************************************************************************
-- IMPORTANT-READ THESE TERMS CAREFULLY BEFORE UTILIZING THE INFORMATION
-- CONTAINED IN THIS APPLICATION NOTE. 
-- NO SUPPORT/NO WARRANTY. THE DESIGN FILES AND DOCUMENTATION ARE PROVIDED
-- WITHOUT SUPPORT OR WARRANTY OF ANY KIND. ACTEL IS NOT OBLIGATED TO 
-- PROVIDE UPDATES, BUG FIXES, OR TECHNICAL SUPPORT AND DISCLAIMS ALL 
-- WARRANTIES INCLUDING, WITHOUT LIMITATION, THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE, AND WARRANTIES OF NON-INFRINGEMENT OF THE
-- RIGHTS OF THIRD PARTIES (INCLUDING, WITHOUT LIMITATION, RIGHTS UNDER PATENT,
-- COPYRIGHT, TRADE SECRET, OR OTHER INTELLECTUAL PROPERTY RIGHTS).  RECIPIENT
-- ACCEPTS THE APPLICATION NOTE DESIGN FILES AND DOCUMENTATION IN "AS-IS"
-- CONDITION.
--*******************************************************************************/
--***************************************************************************************************
-- Main Characteristics and Requirements:
--  The state machine supports 6 comparator measurement channels (using 12 analog
--  voltage inputs). If more channels are desired, the state machine I/Os 
--  should be modified accordingly. The measurement resolution is set to 12-bit as default but
--  it can be changed using WIDTH generic.
--  Each comparator channel is comprised of two voltage inputs (P and N ports). The
--  state machine does not restrict the sampling sequence of AB but requires that the 'N' voltage
--  of a channel be measured immediately after 'P' voltage of that channel. If it is not, the measurement
--  for this channel will be invalidated.
--***************************************************************************************************
-- Functional Description:
--  The state machine starts off in 'idle' state. In this state all the measurement results are
--  reset to zero. The latest_update output is set to "111". The state machine keeps looking at 
--  the channels measured by AB (monitoring ASSC_DONE and CHNUMBER). Once the AB meausures the P voltage 
--  of any comparator channels, the state machine moves to channel_xP state. At this state the 
--  measurement results are recorded and state machine moves to 'wait_next_measurement' state.
--  The state machine expects the next measurement to be taken from the N port of the same channel. If the
--  next measurement is not from the expected channel, it will go to idle state, invalidating the
--  previous P measurement and also clearing all the output results to signal an error in sampling sequence. 
--  If the correct N voltage is measured after P, the state machine moves to "channel_xN" state in which
--  the results are recorded. This result along with the corresponding P result are entered into the subtractor 
--  and the output of the subtractor is sent to the corresponding Diff channel output of the state machine.
--  Then the state machine will move to "wait_next_channel" state and wait for next Dcomparator measurement.
--****************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
library fusion;


entity COMP_SM is 
generic (WIDTH : integer := 12);
port
 (SYSCLK: in std_logic;    --system clock input
  RESETn: in std_logic;    -- active low asynchronous reset
  DATAVALID: in std_logic; -- driven by ASSC_DONE output of Analog Block
  INIT_DONE: in std_logic; -- driven by INIT_DONE output of Analog Block
  ADC_CHNUMBER: in std_logic_vector(4 downto 0); -- driven by ADC_CHNUMBER of Analog Block
  ADC_RESULT: in std_logic_vector(WIDTH-1 downto 0); ---- driven by ADC_RESULT of Analog Block
  -- 6 Diff measurement channels outputs
  Diff_0,Diff_1, Diff_2, Diff_3, Diff_4, Diff_5: out std_logic_vector(WIDTH-1 downto 0);
  Latest_Update: out std_logic_vector(2 downto 0); -- indicates the Diff measurement channel measured last
  New_Measurement: out std_logic -- active high for one clock period. Indicates outputing a new Diff measurement
  );
 end COMP_SM;

 architecture Calculator of COMP_SM is

-- A fast Bren-Kung Subtractor block by SMARTGEN can replace the behavioral subtractor here
 component Subtract is
  generic (WIDTH_SUB: integer := 12);
  port
   ( DataA : in std_logic_vector(WIDTH_SUB-1 downto 0); DataB : in 
   std_logic_vector(WIDTH_SUB-1 downto 0); Sum : out 
   std_logic_vector(WIDTH_SUB-1 downto 0)) ;
 end component;
 
 --Signal Definitions
 --********************************************************************
 -- Pair of Registers holding the measured P and N voltages of a channel
 Signal R1, R2: std_logic_vector(WIDTH-1 downto 0);
 -- output of the subtractor (R1-R2)
 Signal Result : std_logic_vector(WIDTH-1 downto 0);
 -- Used to create New_Measurement output Flag
 Signal completed, completed_sync : std_logic;
 -- used as enables for R1 and R2
 Signal P, N: std_logic;  
 -- Register holding previously measured analog voltage channel
 Signal prev_channel: std_logic_vector( 4 downto 0);
 -- Flag indicating if corresponding N voltage is measured immediately after P voltage
 Signal valid: std_logic;
 --********************************************************************
 
 -- State machine values (5 states)
 type state_values is (idle, channel_xP, channel_xN, wait_next_result, wait_next_channel);
 signal pres_state, next_state: state_values;
 
 -- Define Channel Number constants to make it easy to identify which channels
 -- are associated with Diff 0 through 5
 constant AV0 : std_logic_vector(4 downto 0) := "00001"; 
 constant AC0 : std_logic_vector(4 downto 0) := "00010"; 
 constant AV1 : std_logic_vector(4 downto 0) := "00100"; 
 constant AC1 : std_logic_vector(4 downto 0) := "00101"; 
 constant AV2 : std_logic_vector(4 downto 0) := "00111"; 
 constant AC2 : std_logic_vector(4 downto 0) := "01000"; 
 constant AV3 : std_logic_vector(4 downto 0) := "01010"; 
 constant AC3 : std_logic_vector(4 downto 0) := "01011"; 
 constant AV4 : std_logic_vector(4 downto 0) := "01101"; 
 constant AC4 : std_logic_vector(4 downto 0) := "01110"; 
 constant AV5 : std_logic_vector(4 downto 0) := "10000"; 
 constant AC5 : std_logic_vector(4 downto 0) := "10001"; 
 constant AV6 : std_logic_vector(4 downto 0) := "10011"; 
 constant AC6 : std_logic_vector(4 downto 0) := "10100"; 
 constant AV7 : std_logic_vector(4 downto 0) := "10110"; 
 constant AC7 : std_logic_vector(4 downto 0) := "10111"; 
 constant AV8 : std_logic_vector(4 downto 0) := "11001"; 
 constant AC8 : std_logic_vector(4 downto 0) := "11010"; 
 constant AV9 : std_logic_vector(4 downto 0) := "11100"; 
 constant AC9 : std_logic_vector(4 downto 0) := "11101"; 

 constant Channel0: std_logic_vector(2 downto 0) := "000";
 constant Channel1: std_logic_vector(2 downto 0) := "001";
 constant Channel2: std_logic_vector(2 downto 0) := "010";
 constant Channel3: std_logic_vector(2 downto 0) := "011";
 constant Channel4: std_logic_vector(2 downto 0) := "100";
 constant Channel5: std_logic_vector(2 downto 0) := "101";
 constant no_output: std_logic_vector(2 downto 0) := "111"; 
  
 begin
 
-- assign next state value to presenet state register
 statereg: process (SYSCLK, RESETn, INIT_DONE)
 begin
 if (RESETN = '0' or INIT_DONE = '0') then
    pres_state <= idle;
 elsif (SYSCLK'event and SYSCLK ='1') then
    pres_state <= next_state;
end if;
end process statereg;

-- Body of the state machine defining state transitions 
fsm: process (pres_state, DATAVALID, ADC_CHNUMBER, valid)
begin
case pres_state is
 when idle =>
    case DATAVALID is
      when '1' => 
        case ADC_CHNUMBER is
          when AV0 => next_state <= channel_xP;
          when AV1 => next_state <= channel_xP;
          when AV2 => next_state <= channel_xP;
          when AV3 => next_state <= channel_xP;
          when AV4 => next_state <= channel_xP;
          when AV5 => next_state <= channel_xP;
          when others  => next_state <= idle;
        end case;    
      when others => next_state <= idle; 
     end case;
when channel_xP =>
    next_state <= wait_next_result;
when wait_next_result =>
    case DATAVALID is
      when '1' => 
         case valid is
           when '1' =>  next_state <= channel_xN;
           when others => next_state <= idle;
         end case;  
      when others => next_state <= wait_next_result;
    end case;
when channel_xN => next_state <= wait_next_channel;
when wait_next_channel =>
    case DATAVALID is 
      when '1' => 
        case ADC_CHNUMBER is
          when AV0 => next_state <= channel_xP;
          when AV1 => next_state <= channel_xP;
          when AV2 => next_state <= channel_xP;
          when AV3 => next_state <= channel_xP;
          when AV4 => next_state <= channel_xP;
          when AV5 => next_state <= channel_xP;
          when others  => next_state <= wait_next_channel;
        end case;
     when others => next_state <= wait_next_channel;
     end case;
when others => next_state <= idle;
end case;
end process fsm;

-- output of the state machine is P and N enables to R1 and R2 to record voltage measurements.
-- State machine also outputs a 'completed' signal used for creation of 'New_Measurement' flag.
outputs: process (pres_state)
begin
 case pres_state is
   when channel_xP =>
        P <= '1';
        completed <= '0'; 
	       N <= '0';
   when channel_xN =>
        P <= '0';
        completed <= '1'; 
        N <= '1';
   when others => 
        P <=  '0';
        N <=  '0';
        completed <= '0'; 
 end case;
end process outputs;

-- Assgining output of the subtractor (result) to corresponding Diff channel
process (SYSCLK, RESETn, Completed, ADC_CHNUMBER)
begin 
  if (RESETn = '0') then
    Diff_0 <= (others => '0');
    Diff_1 <= (others => '0');
    Diff_2 <= (others => '0');
    Diff_3 <= (others => '0');
    Diff_4 <= (others => '0');
    Diff_5 <= (others => '0');
	   Latest_Update <= no_output;
  elsif (SYSCLK'event and SYSCLK ='1') then
     if (Completed_sync = '1') then
        case ADC_CHNUMBER is
           when AC0 =>
              Diff_0 <= Result;
              Latest_Update <= Channel0;
           when AC1 =>
              Diff_1 <= Result;
              Latest_Update <= Channel1;
           when AC2 =>
              Diff_2 <= Result;
              Latest_Update <= Channel2;
           when AC3 => 
              Diff_3 <= Result;
              Latest_Update <= Channel3;
           when AC4 => 
              Diff_4 <= Result;
              Latest_Update <= Channel4;
           when AC5 =>
              Diff_5 <= Result;
              Latest_Update <= Channel5;
           when others => NULL;
        end case;
      end if; 
  end if;
end process;

-- asynchrnous subtractor
Subtract_inst: Subtract 
 generic map (12)
 port map
 ( DataA => R1,
   DataB => R2,
   Sum => Result
   ) ;

-- Recording P voltage measurements into R1 register
process (SYSCLK, RESETn, P) begin
 if (RESETn = '0') then
   R1 <= (others => '0');
 elsif (SYSCLK'event and SYSCLK ='1') then
   if (P = '1') then
    R1 <= ADC_RESULT;
   end if;
 end if;
end process;

-- Recording N voltage measurements into R2 register
process (SYSCLK, RESETn, N) begin
 if (RESETn = '0') then
   R2 <= (others => '0');
 elsif (SYSCLK'event and SYSCLK ='1') then
   if (N = '1') then
   R2 <= ADC_RESULT;
   end if;
 end if;
end process;

-- generating New measurement through two back to back registers
process (SYSCLK, RESETn) begin
 if (RESETn = '0') then
   Completed_sync <= '0';
 elsif (SYSCLK'event and SYSCLK ='1') then
    Completed_sync <= Completed;
 end if;
end process;
    
process (SYSCLK, RESETn) begin
 if (RESETn = '0') then
   New_Measurement <= '0';
 elsif (SYSCLK'event and SYSCLK ='1') then
    New_Measurement <= Completed_sync;
 end if;
end process;

-- Recording previously measured analog channel
process (SYSCLK, RESETn) begin
 if (RESETn = '0') then
   prev_channel <= AV0;
 elsif (SYSCLK'event and SYSCLK ='1') then
   if (DATAVALID = '1') then  
    prev_channel <= ADC_CHNUMBER;
   end if; 
 end if;
end process;

-- Determine of the correct N voltage channel is measured after P voltage measurement, validating analog block's sequence
process (prev_channel, ADC_CHNUMBER, DATAVALID, RESETn, INIT_DONE) begin
if (RESETn = '0' or INIT_DONE = '0') then 
  valid <= '0';
elsif (DATAVALID = '1') then 
  case prev_channel is 
    when AV0 => 
      case ADC_CHNUMBER is
        when AC0 => valid <= '1';
        when others => valid <= '0';
      end case;  
    when AV1 => 
      case ADC_CHNUMBER is
        when AC1 => valid <= '1';
        when others => valid <= '0';
      end case;  
    when AV2 =>
      case ADC_CHNUMBER is
        when AC2 => valid <= '1';
        when others => valid <= '0';
      end case;  
    when AV3 =>
      case ADC_CHNUMBER is
        when AC3 => valid <= '1';
        when others => valid <= '0';
      end case;  
    when AV4 =>
      case ADC_CHNUMBER is
        when AC4 => valid <= '1';
        when others => valid <= '0';
      end case;  
    when AV5 =>
      case ADC_CHNUMBER is
        when AC5 => valid <= '1';
        when others => valid <= '0';
      end case;  
    when others  => valid <= '0';
 end case;
end if;
end process;
   


end Calculator;
